import 'dart:io';

import 'package:args/args.dart';
import 'package:critical_test/src/process_output.dart';
import 'package:critical_test/src/run.dart';
import 'package:critical_test/src/unit_tests/failed_tracker.dart';

import 'package:dcli/dcli.dart' hide run;

import 'exceptions/critical_test_exception.dart';
import 'unit_tests/unit_test.dart';

void run(List<String> args) {
  try {
    var processor = ProcessOutput();
    CriticalTest.run(args, processor);

    if (!processor.allPassed) {
      exit(1);
    }
    if (processor.nothingRan) {
      exit(5);
    }
  } on CriticalTestException catch (e) {
    printerr(e.message);
    exit(1);
  }

  exit(0);
}

class CriticalTest {
  static void run(List<String> args, ProcessOutput processor) {
    final parser = ArgParser()
      ..addFlag(
        'help',
        abbr: 'h',
        negatable: false,
        help: 'Shows this usage message.',
      )
      ..addOption('plain-name', abbr: 'N', help: 'Run a unit test by name.')
      ..addFlag('runfailed',
          abbr: 'f',
          negatable: false,
          help:
              'Re-runs only those tests that failed during the last run of critical_test.')
      ..addOption('tags',
          abbr: 't',
          help:
              'Select  unit tests to run via their tags. The syntax must confirm to the --tags option in the test package.')
      ..addOption('exclude-tags',
          abbr: 'x',
          help:
              'Select unit tests to exclude via their tags. The syntax must confirm to the --exclude-tags option in the test package.')
      ..addFlag(
        'show',
        negatable: false,
        abbr: 's',
        help: 'Also show output from successful unit tests.',
      )
      ..addFlag(
        'progress',
        negatable: true,
        abbr: 'p',
        defaultsTo: true,
        help:
            'Show progress messages. Use --no-progress when running with a CI pipeline to minimize noise.',
      )
      ..addFlag(
        'coverage',
        negatable: false,
        defaultsTo: false,
        abbr: 'c',
        help: "Generates test coverage reports in the 'coverage' directory.",
      )
      ..addOption('logPath',
          abbr: 'l',
          help: 'Path to log all output inot. '
              'If set, all tests are logged to the given path.\n'
              'If not set, then all tests are logged to ${Directory.systemTemp.path}/critical_test/unit_test.log')
      ..addFlag(
        'hooks',
        abbr: 'o',
        negatable: true,
        defaultsTo: false,
        help: 'Supresses running of the pre and post hooks.',
      )
      ..addFlag(
        'warmup',
        abbr: 'w',
        defaultsTo: true,
        negatable: true,
        help: '''
Causes pub get to be run on all pubspec.yaml files found in the package.
Unit tests will fail if pub get hasn't been run.''',
      )
      ..addFlag(
        'track',
        abbr: 'k',
        hide: true,
        help: 'Used to force the recording of failures in .failed_tracker.',
      )
      ..addOption(
        'tracker',
        defaultsTo: FailedTracker.defaultFilename,
        hide: true,
        help:
            'Used to define an alternate filename for the fail test tracker. This is intended only for internal testing',
      )
      ..addFlag(
        'verbose',
        negatable: false,
        abbr: 'v',
        hide: true,
        help: 'Verbose logging for debugging of critical test.',
      );

    late final ArgResults parsed;
    try {
      parsed = parser.parse(args);
    } on FormatException catch (e) {
      printerr(red(e.message));
      printerr('');
      showUsage(parser);
    }

    if (parsed['help'] as bool == true) {
      showUsage(parser);
    }

    var verbose = parsed['verbose'] as bool;
    Settings().setVerbose(enabled: verbose);

    processor.showSuccess = parsed['show'] as bool;
    processor.showProgress = parsed['progress'] as bool;

    var coverage = parsed['coverage'] as bool;
    var warmup = parsed['warmup'] as bool;
    var track = parsed['track'] as bool;
    var hooks = parsed['hooks'] as bool;

    var runFailed = parsed['runfailed'] as bool;

    if (!atMostOne([runFailed, parsed.wasParsed('plain-name')])) {
      printerr(red('You may only pass one of --plain-name or --runfailed'));
      showUsage(parser);
    }

    final trackerFilename = parsed['tracker'] as String;

    if (parsed.wasParsed('plain-name') &&
        (parsed.wasParsed('tags') || parsed.wasParsed('exclude-tags'))) {
      printerr(red(
          'You cannot combine "--plain-name" with "--tag" or "--exclude-tags"'));
      showUsage(parser);
    }

    if (parsed.wasParsed('logPath')) {
      final _logPath = truepath(parsed['logPath'] as String);
      if (exists(_logPath) && !isFile(_logPath)) {
        printerr(red('--logPath must specify a file'));
        showUsage(parser);
      }
      processor.logPath = _logPath;
    }

    String? tags;
    if (parsed.wasParsed('tags')) {
      tags = parsed['tags'] as String;
    }

    String? excludeTags;
    if (parsed.wasParsed('exclude-tags')) {
      excludeTags = parsed['exclude-tags'] as String;
    }

    final pathToProjectRoot = DartProject.fromPath(pwd).pathToProjectRoot;

    String? testName;

    if (parsed.wasParsed('plain-name')) {
      testName = parsed['plain-name'] as String;
    }

    try {
      if (runFailed) {
        runFailedTests(
            processor: processor,
            pathToProjectRoot: pathToProjectRoot,
            tags: tags,
            excludeTags: excludeTags,
            coverage: coverage,
            warmup: warmup,
            hooks: hooks,
            trackerFilename: trackerFilename);
      } else {
        if (parsed.rest.isEmpty) {
          if (testName == null) {
            runPackageTests(
                processor: processor,
                pathToProjectRoot: pathToProjectRoot,
                tags: tags,
                excludeTags: excludeTags,
                coverage: coverage,
                warmup: warmup,
                hooks: hooks,
                trackerFilename: trackerFilename);
          } else {
            // run named test
            runSingleTest(
                processor: processor,
                unitTest: UnitTest(pathTo: null, testName: testName),
                pathToProjectRoot: pathToProjectRoot,
                tags: tags,
                excludeTags: excludeTags,
                coverage: coverage,
                warmup: warmup,
                track: track,
                hooks: hooks,
                trackerFilename: trackerFilename);
          }
        } else {
          /// Process each director or library passed.
          for (final dirOrFile in parsed.rest) {
            if (!exists(dirOrFile)) {
              printerr(red("The path ${truepath(dirOrFile)} doesn't exist."));
              showUsage(parser);
            }

            runSingleTest(
                processor: processor,
                unitTest: UnitTest(pathTo: dirOrFile, testName: testName),
                pathToProjectRoot: pathToProjectRoot,
                tags: tags,
                excludeTags: excludeTags,
                coverage: coverage,
                warmup: warmup,
                track: track,
                hooks: hooks,
                trackerFilename: trackerFilename);
          }
        }
      }

      if (processor.nothingRan) {
        print(orange('No tests ran!'));
      } else if (processor.allPassed) {
        print(green(
            'All tests passed. Success: ${processor.successCount}, Skipped: ${processor.skippedCount}'));
      } else {
        printerr(
            '${red('Some tests failed!')} Errors: ${red('${processor.erorrCount}')}, Success: ${green('${processor.successCount}')}, Skipped: ${blue('${processor.skippedCount}')}');
      }
    } on CriticalTestException catch (e) {
      printerr('A non recoverable error occured: ${e.message}');
    }
  }

  /// Show useage.
  static void showUsage(ArgParser parser) {
    print(orange('Usage: critical_test [arguments]'));
    print(
        'Runs unit tests only showing output from failed tests and allows you to just re-run failed tests.');
    print(blue('Run all tests'));
    print('critical_test');
    print('');
    print(blue('Re-run failed tests'));
    print('critical_tests --runfailed');
    print('');
    print(blue('Run all tests in a Dart Library or directory'));
    print('critical_tests <path to test>');
    print('');
    print(blue('Run a single test by name'));
    print(
        'critical_tests --plain-name="[<group name> <group name> ...] <test name>"');
    print('');
    print(parser.usage);
    exit(1);
  }

  /// no more than one of the passed bools may be true
  static bool atMostOne(List<bool> list) {
    var count = 0;

    for (final val in list) {
      if (val) count++;
    }
    return count <= 1;
  }
}
